const fs        = require('fs');
const path      = require('path');
const assert    = require('assert');
const IniParser = require('./iniparser.js');


module.exports = class DocorSetting {
    constructor(ini_file_path) {
        this._root_path      = '';
        this._setting_library_list   = []; /*<this._library_struct>*/
        this._setting_project_title  = '';

        this._library_struct = {
            '_name'              : '',
            '_namespace'         : '',
            '_abs_path'          : '',
            '_modules'           : [] /*<this._module_struct>*/
        };

        this._module_struct  = {
            '_name'              : '',
            '_regex'             : '',
            '_abs_path'          : '',
            '_main_module'       : {
                '_sources'       : [],
                '_headers'       : []
            },
            '_sub_module'        : {
                '_sources'       : []
            }
        };

        this._parse(ini_file_path);
    }

    get rootPath() {
        return this._root_path;
    }

    get title() {
        return this._setting_project_title;
    }

    _parse(ini_file_path) {
        const abs_ini_file_path = path.resolve(ini_file_path);
        if (!abs_ini_file_path ||! fs.existsSync(abs_ini_file_path)) {
            return false;
        }

        let ini_setting = this._loadIniConfig(abs_ini_file_path);
        if (!ini_setting) {
            return false;
        }

        if ((!ini_setting.Config)||
            (!ini_setting.Config.ProjectFolder)||
            (!ini_setting.Config.LibrarySections)) {
            return false;
        }

        // Keep Project Title
        let project_title = ini_setting.Config.ProjectTitle;
        if (project_title) {
            this._setting_project_title = project_title.trim();
        }

        // Create settings for library
        let root_path = ini_setting.Config.ProjectFolder.trim();
        if (!path.isAbsolute(root_path)) {
            const file_path = path.join(path.dirname(ini_file_path), root_path);
            root_path = path.resolve(file_path);
        }


        //
        // Handle library section
        //
        let sect_strings = ini_setting.Config.LibrarySections.replace(/\s/g, '');
        let lib_name_str_list = sect_strings.split(',');
        if (lib_name_str_list && lib_name_str_list.length > 0) {
            for (let idx = 0; idx < lib_name_str_list.length; idx++) {

                let setting_library        = JSON.parse(JSON.stringify(this._library_struct));
                setting_library._namespace = lib_name_str_list[idx];

                if (setting_library._namespace &&
                    ini_setting[setting_library._namespace].Name &&
                    ini_setting[setting_library._namespace].Folder) {

                    setting_library._name = ini_setting[setting_library._namespace].Name.trim();

                    const dir_path = ini_setting[setting_library._namespace].Folder;
                    setting_library._abs_path = path.resolve(path.join(root_path, dir_path));
                    this._setting_library_list.push(setting_library);
                }
            }
        }


        //
        // Create settings for modules in every library.
        //
        for (let lib_idx = 0; lib_idx < this._setting_library_list.length; lib_idx++) {
            let setting_library = this._setting_library_list[lib_idx];
            let found_ini_section = this._findSectionByName(ini_setting, setting_library._namespace);
            if (found_ini_section) {
                const library_folder_name = found_ini_section.Folder;
                const library_folder_path = path.resolve(path.join(root_path, library_folder_name));
                for (let key_name in found_ini_section) {

                    // Skip key name is `Folder` or `Name`.
                    if (('Folder' === key_name) ||
                        ('Name'   === key_name)) {
                        continue;
                    }

                    const key_value = found_ini_section[key_name];
                    const formatted_key_value = key_value.replace(/\s/g, '');
                    let new_setting_module = this._createSettingForModule(library_folder_path, key_name, formatted_key_value);
                    if (new_setting_module) {
                        setting_library._modules.push(new_setting_module);
                    }
                }
            }
        }

        return true;
    }

    _createSettingForModule(module_folder_path, module_name, module_regex_file_name) {
        if (!module_folder_path instanceof String ||
            !module_name instanceof String ||
            !module_regex_file_name instanceof String) {
            return null;
        }

        const abs_module_folder_path = path.resolve(module_folder_path);
        if (!fs.existsSync(abs_module_folder_path)) {
            return null;
        }

        let new_setting_module = JSON.parse(JSON.stringify(this._module_struct));
        if (!new_setting_module) {
            return null;
        }

        const file_list = fs.readdirSync(abs_module_folder_path);
        if (file_list && file_list.length > 0) {
            new_setting_module._name     = module_name;
            new_setting_module._regex    = module_regex_file_name;
            new_setting_module._abs_path = path.resolve(abs_module_folder_path);

            for (let idx = 0; idx < file_list.length; idx++) {
                const regex_result = file_list[idx].match(module_regex_file_name);
                if (regex_result && regex_result.length >= 2) {
                    if (regex_result[1]) {
                        const filename = regex_result[1].toLowerCase();
                        if (filename.endsWith('.h')) {
                            new_setting_module._main_module._headers.push(regex_result[1]);
                        } else if (filename.endsWith('.c')) {
                            new_setting_module._main_module._sources.push(regex_result[1]);
                        } else {
                            assert(false);
                        }
                    } else if (regex_result[2]) {
                        new_setting_module._sub_module._sources.push(regex_result[2]);
                    } else {
                        assert(false);
                    }
                }
            }
        }
        return new_setting_module;
    }

    _findSectionByName(ini_config, lib_name) {
        if (!ini_config instanceof  Object) {
            return null;
        }
        if (!lib_name instanceof String) {
            return null;
        }

        let sect_value = null;
        for (const sect_name in ini_config) {
            if (sect_name === lib_name) {
                sect_value = ini_config[lib_name];
                break;
            }
        }

        return sect_value;
    }

    get rootPath() {
        return this._root_path;
    }

    get libraryList() {
        return this._setting_library_list;
    }


    _loadIniConfig(config_filepath) {
        let ini_setting = IniParser.parseSync(config_filepath);

        if (ini_setting) {
            for (let sect_name in ini_setting) {
                let sect_value = ini_setting[sect_name];
                for (let key_name in sect_value) {
                    const key_value = sect_value[key_name];
                    if (('Name'   !== key_name) &&
                        ('Folder' !== key_name)) {
                        sect_value[key_name] = key_value.replace(/\s/g, '');
                    }
                }
            }
            return ini_setting;
        } else {
            return [];
        }
    }
};