/*!
 * docor
 * Copyright(c) 2018 Douglas Chen <dougpuob@gmail.com>
 * MIT Licensed
 */
'use strict';

const fs   = require('fs');
const path = require('path');
const assert = require('assert');

const CAstJson       = require('node-castjson');
const PaJSDoc        = require('node-pajsdoc');
const DocorLibrary   = require('./docor-library');
const DocorModule    = require('./docor-module');
const DocorSubModule = require('./docor-submodule');
const DocorFunction  = require('./docor-function');
const DocorParameter = require('./docor-parameter');
const DocorSetting   = require('./docor-setting');

module.exports = class Docor {
    constructor(config_file_path) {
        this._docor_library_list /*<DocorLibrary>*/ = [];
        this._docor_setting_list            = undefined;
        this._docor_setting_project_title   = '';

        if (config_file_path) {
            let docorSetting = new DocorSetting(config_file_path);
            if (docorSetting) {
                this._docor_setting_list          = docorSetting.libraryList;
                this._docor_setting_project_title = docorSetting.title;
            }

            if (this._docor_setting_list) {
                this._docor_library_list = this._classify(this._docor_setting_list);
            }
        }
    }

    get libraryList() {
        return this._docor_library_list;
    }
    get settingList() {
        return this._docor_setting_list;
    }
    get projectTitle() {
        return this._docor_setting_project_title;
    }


    _callCastJson(module_filepath) {
        let output_jsonbuf = {
            data: null
        }

        if (fs.existsSync(module_filepath)) {
            let castjson = new CAstJson();            
            castjson.parse(module_filepath, output_jsonbuf);
            if (output_jsonbuf.data) {
                return JSON.parse(output_jsonbuf.data);    
            }
        }
        
        return {
            'File'      : module_filepath,
            'Functions' : []
        };        
    }


    _classify(docor_setting_list) {
        if (!docor_setting_list || !Array.isArray(docor_setting_list)) {
            return [];
        }

        let bResult = false;
        let docor_library_list = [];
        for (let idx_lib = 0; idx_lib < docor_setting_list.length; idx_lib++) {
            const docor_setting = docor_setting_list[idx_lib];
            if (!docor_setting) {
                break;
            }

            let new_docor_library = new DocorLibrary();
            new_docor_library.Name      = docor_setting._name;
            new_docor_library.Namespace = docor_setting._namespace;

            for (let idx_mod = 0; idx_mod < docor_setting._modules.length; idx_mod++) {
                const setting_module = docor_setting._modules[idx_mod];
                if ((!setting_module) ||
                    (!setting_module._main_module) ||
                    (!setting_module._main_module._sources)||
                    (0 === setting_module._main_module._sources.length)) {
                    continue;
                }


                if (setting_module) {
                    //
                    // Create major module object.
                    //
                    let castjson                 = new CAstJson();
                    let new_docor_module         = new DocorModule();
                    new_docor_module.Name        = setting_module._name;
                    new_docor_module.absPath     = setting_module._abs_path;
                    new_docor_module.File.Source = setting_module._main_module._sources;
                    new_docor_module.File.Header = setting_module._main_module._headers;

                    // Create castjson object then append AST data to current new docor module.
                    let final_json_obj = null;
                    for (let idx = 0; idx < new_docor_module.File.Source.length; idx++) {
                        let json_buf = {};
                        let file_path = path.resolve(path.join(setting_module._abs_path, new_docor_module.File.Source[idx]));

                        console.log(file_path);

                        if (null === final_json_obj) {
                            final_json_obj = this._callCastJson(file_path);
                        } else {
                            const temp_json_obj = this._callCastJson(file_path);
                            final_json_obj.Functions = final_json_obj.Functions.concat(temp_json_obj.Functions);
                        }
                    }
                    new_docor_module._astjson = final_json_obj;
                    assert(new_docor_module._astjson);

                    // Create JSDoc object(PAJSDoc) then append JSDoc data to current new docor module.
                    let merged_relative_files = new_docor_module.File.Source.concat(new_docor_module.File.Header);
                    merged_relative_files.forEach(function(item, index, array) { array[index] = path.resolve(path.join(setting_module._abs_path, array[index]))});
                    new_docor_module._pajsdoc = new PaJSDoc(merged_relative_files);
                    assert(new_docor_module._pajsdoc);


                    //
                    // Create sub module object.
                    //
                    for (let idx = 0; idx < setting_module._sub_module._sources.length; idx++) {
                        let new_docor_submod_obj         = new DocorSubModule();
                        new_docor_submod_obj.absPath     = setting_module._abs_path;
                        new_docor_submod_obj.File.Source = setting_module._sub_module._sources[idx];

                        let file_path = path.resolve(path.join(setting_module._abs_path, setting_module._sub_module._sources[idx]));

                        new_docor_submod_obj._pajsdoc = new PaJSDoc(file_path);
                        new_docor_submod_obj._astjson = this._callCastJson(file_path);
                        new_docor_module.Submodules.push(new_docor_submod_obj);
                    }


                    //
                    // Create function objects for modules.
                    //
                    bResult = this._createFunctionObjectList(new_docor_module);
                    if (bResult) {
                        // Create function objects for submodules.
                        for (let idx_submod = 0; idx_submod < new_docor_module.Submodules.length; idx_submod++) {
                            this._createFunctionObjectList(new_docor_module.Submodules[idx_submod]);
                        }
                        new_docor_library.Modules.push(new_docor_module);
                    }
                }
            }
            if (bResult) {
                docor_library_list.push(new_docor_library)
            }
        }
        return docor_library_list;
    }

    _createFunctionObjectList(docor_module) {
        if (!docor_module instanceof DocorModule) {
            return false;
        }

        if (!docor_module._astjson) {
            return false;
        }

        for (let func_json_item in docor_module._astjson.Functions) {
            let func_json_obj = docor_module._astjson.Functions[func_json_item];
            if (!func_json_obj instanceof Object) {
                continue;
            }

            let new_docor_function = new DocorFunction();
            new_docor_function.Name   = func_json_obj.Name;
            new_docor_function.Return = func_json_obj.Return.replace(/\s/g, '');

            if (func_json_obj.Parameters) {
                for (let param_json_item in func_json_obj.Parameters) {
                    const param_name = param_json_item;
                    const param_type = func_json_obj.Parameters[param_json_item];

                    let new_docor_parameter = new DocorParameter();
                    new_docor_parameter.Name = param_name;
                    new_docor_parameter.Type = param_type;

                    new_docor_function.Parameters.push(new_docor_parameter);
                }
            }

            docor_module._all_functions.push(new_docor_function);
        }

        const bStatus = this._fillFunctionInfoFromJSDoc(docor_module) &&
                        this._fillModuleInfoFromJSDoc(docor_module);

        return bStatus;
    }

    _fillModuleInfo(module_object, jsdoc_block) {
        if (jsdoc_block && jsdoc_block instanceof Object && (module_object && module_object instanceof Object)) {
            if (jsdoc_block.namespace) {
                module_object.Namespace = jsdoc_block.namespace.Name;
            }
            if (jsdoc_block.author) {
                module_object.Author = jsdoc_block.author.Name;
            }
            if (jsdoc_block.member) {
                module_object.Member = jsdoc_block.member.Type;
            }
            if (jsdoc_block.module) {
                module_object.Module = jsdoc_block.module.Name;
            }
            return true;
        } else {
            return false;
        }
    }

    _fillModuleInfoFromJSDoc(module_object) {
        if (!module_object || !(module_object instanceof DocorModule)) {
            return false;
        }

        // Main module
        const pajsdoc = module_object._pajsdoc;
        if (pajsdoc && pajsdoc instanceof PaJSDoc) {
            let jsdoc_block = pajsdoc.searchByTagTypeName('module');
            this._fillModuleInfo(module_object, jsdoc_block);
        }

        // Submodule
        for (let item in module_object.Submodules) {
            let submodule_obj = module_object.Submodules[item];
            if (submodule_obj instanceof DocorSubModule) {
                const submodule_jsdoc = submodule_obj._pajsdoc;
                let jsdoc_block = submodule_jsdoc.searchByTagTypeName('module');
                this._fillModuleInfo(submodule_obj, jsdoc_block);
            }
        }
        return true;
    }

    _fillFunctionInfoFromJSDoc(docor_module_object) {
        if (!docor_module_object || !(docor_module_object instanceof DocorModule)) {
            return false;
        }

        const pajsdoc = docor_module_object._pajsdoc;
        if (!pajsdoc instanceof  PaJSDoc) {
            return false;
        }

        for (let idx=0;idx < docor_module_object._all_functions.length; idx++) {
            let func_obj = docor_module_object._all_functions[idx];
            if (!func_obj || !(func_obj instanceof DocorFunction)) {
                continue;
            }

            let jsdoc_func_obj = pajsdoc.searchByFunction(func_obj.Name, 'function');
            if (jsdoc_func_obj && jsdoc_func_obj instanceof Object) {
                func_obj.JsDoc = jsdoc_func_obj;

                if (jsdoc_func_obj.return)      { func_obj.Return      = jsdoc_func_obj.return.Type;             }
                if (jsdoc_func_obj.parameters)  { func_obj.Parameters  = jsdoc_func_obj.parameters;              }
                if (jsdoc_func_obj.namespace)   { func_obj.Namespace   = jsdoc_func_obj.namespace.Name;          }
                if (jsdoc_func_obj.author)      { func_obj.Author      = jsdoc_func_obj.author.Name;             }
                if (jsdoc_func_obj.summary)     { func_obj.Summary     = jsdoc_func_obj.summary.Description;     }
                if (jsdoc_func_obj.description) { func_obj.Description = jsdoc_func_obj.description.Description; }
            }
        }
        return true;
    }
};